/*
 * Copyright (c) 2010-2011, The MiCode Open Source Community (www.micode.net)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package net.micode.notes.data;

import android.content.ContentValues;
import android.content.Context;
import android.database.sqlite.SQLiteDatabase;
import android.database.sqlite.SQLiteOpenHelper;
import android.util.Log;

import net.micode.notes.data.Notes.DataColumns;
import net.micode.notes.data.Notes.DataConstants;
import net.micode.notes.data.Notes.NoteColumns;


public class NotesDatabaseHelper extends SQLiteOpenHelper {
    // 数据库名称
    private static final String DB_NAME = "note.db";

    // 数据库版本号
    private static final int DB_VERSION = 4;

    // 定义数据库中的表名
    public interface TABLE {
        // 表名：note
        public static final String NOTE = "note";

        // 表名：data
        public static final String DATA = "data";
    }

    // 定义一个常量，用于标识数据库帮助类的标签
    private static final String TAG = "NotesDatabaseHelper";

    // 定义一个静态变量，用于保存数据库帮助类的实例
    private static NotesDatabaseHelper mInstance;

    // 定义一个常量，用于保存创建笔记表的SQL语句
    private static final String CREATE_NOTE_TABLE_SQL =
            "CREATE TABLE " + TABLE.NOTE + "(" +
                    // 定义笔记表的ID列，类型为INTEGER，为主键
                    NoteColumns.ID + " INTEGER PRIMARY KEY," +
                    // 定义笔记表的父ID列，类型为INTEGER，默认值为0
                    NoteColumns.PARENT_ID + " INTEGER NOT NULL DEFAULT 0," +
                    // 定义笔记表的提醒日期列，类型为INTEGER，默认值为0
                    NoteColumns.ALERTED_DATE + " INTEGER NOT NULL DEFAULT 0," +
                    // 定义笔记表的背景颜色ID列，类型为INTEGER，默认值为0
                    NoteColumns.BG_COLOR_ID + " INTEGER NOT NULL DEFAULT 0," +
                    // 定义笔记表的创建日期列，类型为INTEGER，默认值为当前时间戳
                    NoteColumns.CREATED_DATE + " INTEGER NOT NULL DEFAULT (strftime('%s','now') * 1000)," +
                    // 定义笔记表是否有附件列，类型为INTEGER，默认值为0
                    NoteColumns.HAS_ATTACHMENT + " INTEGER NOT NULL DEFAULT 0," +
                    // 定义笔记表的修改日期列，类型为INTEGER，默认值为当前时间戳
                    NoteColumns.MODIFIED_DATE + " INTEGER NOT NULL DEFAULT (strftime('%s','now') * 1000)," +
                    // 定义笔记表的笔记数量列，类型为INTEGER，默认值为0
                    NoteColumns.NOTES_COUNT + " INTEGER NOT NULL DEFAULT 0," +
                    // 定义笔记表的摘要列，类型为TEXT，默认值为空字符串
                    NoteColumns.SNIPPET + " TEXT NOT NULL DEFAULT ''," +
                    // 定义笔记表的类型列，类型为INTEGER，默认值为0
                    NoteColumns.TYPE + " INTEGER NOT NULL DEFAULT 0," +
                    // 定义笔记表的Widget ID列，类型为INTEGER，默认值为0
                    NoteColumns.WIDGET_ID + " INTEGER NOT NULL DEFAULT 0," +
                    // 定义笔记表的Widget类型列，类型为INTEGER，默认值为-1
                    NoteColumns.WIDGET_TYPE + " INTEGER NOT NULL DEFAULT -1," +
                    // 定义笔记表的同步ID列，类型为INTEGER，默认值为0
                    NoteColumns.SYNC_ID + " INTEGER NOT NULL DEFAULT 0," +
                    // 定义笔记表的本地修改列，类型为INTEGER，默认值为0
                    NoteColumns.LOCAL_MODIFIED + " INTEGER NOT NULL DEFAULT 0," +
                    // 定义笔记表的原始父ID列，类型为INTEGER，默认值为0
                    NoteColumns.ORIGIN_PARENT_ID + " INTEGER NOT NULL DEFAULT 0," +
                    // 定义笔记表的GTask ID列，类型为TEXT，默认值为空字符串
                    NoteColumns.GTASK_ID + " TEXT NOT NULL DEFAULT ''," +
                    // 定义笔记表的版本列，类型为INTEGER，默认值为0
                    NoteColumns.VERSION + " INTEGER NOT NULL DEFAULT 0" +
                    ")";

    // 定义创建数据表的SQL语句
    private static final String CREATE_DATA_TABLE_SQL =
            "CREATE TABLE " + TABLE.DATA + "(" +
                    // 定义ID列，类型为INTEGER，为主键
                    DataColumns.ID + " INTEGER PRIMARY KEY," +
                    // 定义MIME_TYPE列，类型为TEXT，不能为空
                    DataColumns.MIME_TYPE + " TEXT NOT NULL," +
                    // 定义NOTE_ID列，类型为INTEGER，不能为空，默认值为0
                    DataColumns.NOTE_ID + " INTEGER NOT NULL DEFAULT 0," +
                    // 定义CREATED_DATE列，类型为INTEGER，不能为空，默认值为当前时间戳
                    NoteColumns.CREATED_DATE + " INTEGER NOT NULL DEFAULT (strftime('%s','now') * 1000)," +
                    // 定义MODIFIED_DATE列，类型为INTEGER，不能为空，默认值为当前时间戳
                    NoteColumns.MODIFIED_DATE + " INTEGER NOT NULL DEFAULT (strftime('%s','now') * 1000)," +
                    // 定义CONTENT列，类型为TEXT，不能为空，默认值为空字符串
                    DataColumns.CONTENT + " TEXT NOT NULL DEFAULT ''," +
                    // 定义DATA1列，类型为INTEGER
                    DataColumns.DATA1 + " INTEGER," +
                    // 定义DATA2列，类型为INTEGER
                    DataColumns.DATA2 + " INTEGER," +
                    // 定义DATA3列，类型为TEXT，不能为空，默认值为空字符串
                    DataColumns.DATA3 + " TEXT NOT NULL DEFAULT ''," +
                    // 定义DATA4列，类型为TEXT，不能为空，默认值为空字符串
                    DataColumns.DATA4 + " TEXT NOT NULL DEFAULT ''," +
                    // 定义DATA5列，类型为TEXT，不能为空，默认值为空字符串
                    DataColumns.DATA5 + " TEXT NOT NULL DEFAULT ''" +
                    ")";

    // 创建一个名为note_id_index的索引，用于在DATA表中根据NOTE_ID进行查询
    private static final String CREATE_DATA_NOTE_ID_INDEX_SQL =
            "CREATE INDEX IF NOT EXISTS note_id_index ON " +
                    TABLE.DATA + "(" + DataColumns.NOTE_ID + ");";

    /**
     * Increase folder's note count when move note to the folder
     */
    // 创建一个名为increase_folder_count_on_update的触发器，用于在NOTE表中更新PARENT_ID时，增加对应文件夹的note count
    // 创建一个触发器，在更新Note表的PARENT_ID字段后，增加对应文件夹的NOTES_COUNT字段
    private static final String NOTE_INCREASE_FOLDER_COUNT_ON_UPDATE_TRIGGER =
            "CREATE TRIGGER increase_folder_count_on_update "+
                    " AFTER UPDATE OF " + NoteColumns.PARENT_ID + " ON " + TABLE.NOTE +
                    " BEGIN " +
                    "  UPDATE " + TABLE.NOTE +
                    "   SET " + NoteColumns.NOTES_COUNT + "=" + NoteColumns.NOTES_COUNT + " + 1" +
                    "  WHERE " + NoteColumns.ID + "=new." + NoteColumns.PARENT_ID + ";" +
                    " END";

    /**
     * Decrease folder's note count when move note from folder
     */
    // 创建一个触发器，在更新Note表的PARENT_ID字段后，减少对应文件夹的NOTES_COUNT字段
    // 创建一个触发器，在更新Note表的PARENT_ID字段后，减少对应文件夹的NOTES_COUNT字段
    private static final String NOTE_DECREASE_FOLDER_COUNT_ON_UPDATE_TRIGGER =
            "CREATE TRIGGER decrease_folder_count_on_update " +
                    " AFTER UPDATE OF " + NoteColumns.PARENT_ID + " ON " + TABLE.NOTE +
                    " BEGIN " +
                    "  UPDATE " + TABLE.NOTE +
                    "   SET " + NoteColumns.NOTES_COUNT + "=" + NoteColumns.NOTES_COUNT + "-1" +
                    "  WHERE " + NoteColumns.ID + "=old." + NoteColumns.PARENT_ID +
                    "  AND " + NoteColumns.NOTES_COUNT + ">0" + ";" +
                    " END";

    /**
     * Increase folder's note count when insert new note to the folder
     */
    // 创建一个触发器，在插入笔记后，增加父文件夹的笔记数量
    private static final String NOTE_INCREASE_FOLDER_COUNT_ON_INSERT_TRIGGER =
            "CREATE TRIGGER increase_folder_count_on_insert " +
                    " AFTER INSERT ON " + TABLE.NOTE +
                    " BEGIN " +
                    "  UPDATE " + TABLE.NOTE +
                    "   SET " + NoteColumns.NOTES_COUNT + "=" + NoteColumns.NOTES_COUNT + " + 1" +
                    "  WHERE " + NoteColumns.ID + "=new." + NoteColumns.PARENT_ID + ";" +
                    " END";

    /**
     * Decrease folder's note count when delete note from the folder
     */
    // 创建一个触发器，在删除笔记后，减少父文件夹中的笔记数量
    private static final String NOTE_DECREASE_FOLDER_COUNT_ON_DELETE_TRIGGER =
            "CREATE TRIGGER decrease_folder_count_on_delete " +
                    " AFTER DELETE ON " + TABLE.NOTE +
                    " BEGIN " +
                    "  UPDATE " + TABLE.NOTE +
                    "   SET " + NoteColumns.NOTES_COUNT + "=" + NoteColumns.NOTES_COUNT + "-1" +
                    "  WHERE " + NoteColumns.ID + "=old." + NoteColumns.PARENT_ID +
                    "  AND " + NoteColumns.NOTES_COUNT + ">0;" +
                    " END";

    /**
     * Update note's content when insert data with type {@link DataConstants#NOTE}
     */
    // 创建一个触发器，在插入数据后更新笔记内容
    private static final String DATA_UPDATE_NOTE_CONTENT_ON_INSERT_TRIGGER =
            "CREATE TRIGGER update_note_content_on_insert " +
                    " AFTER INSERT ON " + TABLE.DATA + // 在数据表上创建触发器
                    " WHEN new." + DataColumns.MIME_TYPE + "='" + DataConstants.NOTE + "'" + // 当插入的数据类型为笔记时
                    " BEGIN" +
                    "  UPDATE " + TABLE.NOTE + // 更新笔记表
                    "   SET " + NoteColumns.SNIPPET + "=new." + DataColumns.CONTENT + // 将新插入的数据内容设置为笔记的片段
                    "  WHERE " + NoteColumns.ID + "=new." + DataColumns.NOTE_ID + ";" + // 根据笔记ID更新笔记表
                    " END";

    /**
     * Update note's content when data with {@link DataConstants#NOTE} type has changed
     */
    // 创建一个名为update_note_content_on_update的触发器，在更新DATA表后执行
    private static final String DATA_UPDATE_NOTE_CONTENT_ON_UPDATE_TRIGGER =
            "CREATE TRIGGER update_note_content_on_update " +
                    " AFTER UPDATE ON " + TABLE.DATA +
                    " WHEN old." + DataColumns.MIME_TYPE + "='" + DataConstants.NOTE + "'" +
                    " BEGIN" +
                    "  UPDATE " + TABLE.NOTE +
                    "   SET " + NoteColumns.SNIPPET + "=new." + DataColumns.CONTENT +
                    "  WHERE " + NoteColumns.ID + "=new." + DataColumns.NOTE_ID + ";" +
                    " END";

    /**
     * Update note's content when data with {@link DataConstants#NOTE} type has deleted
     */
    private static final String DATA_UPDATE_NOTE_CONTENT_ON_DELETE_TRIGGER =
            "CREATE TRIGGER update_note_content_on_delete " +
                    " AFTER delete ON " + TABLE.DATA +
                    " WHEN old." + DataColumns.MIME_TYPE + "='" + DataConstants.NOTE + "'" +
                    " BEGIN" +
                    "  UPDATE " + TABLE.NOTE +
                    "   SET " + NoteColumns.SNIPPET + "=''" +
                    "  WHERE " + NoteColumns.ID + "=old." + DataColumns.NOTE_ID + ";" +
                    " END";

    /**
     * Delete datas belong to note which has been deleted
     */
    // 创建一个名为delete_data_on_delete的触发器，在删除note表中的数据后，删除data表中的数据
    private static final String NOTE_DELETE_DATA_ON_DELETE_TRIGGER =
            "CREATE TRIGGER delete_data_on_delete " +
                    " AFTER DELETE ON " + TABLE.NOTE +
                    " BEGIN" +
                    "  DELETE FROM " + TABLE.DATA +
                    "   WHERE " + DataColumns.NOTE_ID + "=old." + NoteColumns.ID + ";" +
                    " END";

    /**
     * Delete notes belong to folder which has been deleted
     */
    // 创建一个触发器，在删除笔记时，删除其子笔记
    private static final String FOLDER_DELETE_NOTES_ON_DELETE_TRIGGER =
            "CREATE TRIGGER folder_delete_notes_on_delete " +
                    " AFTER DELETE ON " + TABLE.NOTE +
                    " BEGIN" +
                    "  DELETE FROM " + TABLE.NOTE +
                    "   WHERE " + NoteColumns.PARENT_ID + "=old." + NoteColumns.ID + ";" +
                    " END";

    /**
     * Move notes belong to folder which has been moved to trash folder
     */
    // 创建一个触发器，当更新NOTE表时，如果新父ID等于垃圾文件夹ID，则更新NOTE表中的父ID为垃圾文件夹ID
    private static final String FOLDER_MOVE_NOTES_ON_TRASH_TRIGGER =
            "CREATE TRIGGER folder_move_notes_on_trash " +
                    " AFTER UPDATE ON " + TABLE.NOTE +
                    " WHEN new." + NoteColumns.PARENT_ID + "=" + Notes.ID_TRASH_FOLER +
                    " BEGIN" +
                    "  UPDATE " + TABLE.NOTE +
                    "   SET " + NoteColumns.PARENT_ID + "=" + Notes.ID_TRASH_FOLER +
                    "  WHERE " + NoteColumns.PARENT_ID + "=old." + NoteColumns.ID + ";" +
                    " END";

    public NotesDatabaseHelper(Context context) {
        super(context, DB_NAME, null, DB_VERSION);
    }

    /**
     * 创建note表
     */
    public void createNoteTable(SQLiteDatabase db) {
        // 执行创建note表的SQL语句
        db.execSQL(CREATE_NOTE_TABLE_SQL);
        // 重新创建note表的触发器
        reCreateNoteTableTriggers(db);
        // 创建系统文件夹
        createSystemFolder(db);
        // 打印日志，note表已创建
        Log.d(TAG, "note table has been created");
    }

    /**
     * 重新创建note表的触发器
     */
    // 重新创建Note表触发器
    private void reCreateNoteTableTriggers(SQLiteDatabase db) {
        // 删除increase_folder_count_on_update触发器
        db.execSQL("DROP TRIGGER IF EXISTS increase_folder_count_on_update");
        // 删除decrease_folder_count_on_update触发器
        db.execSQL("DROP TRIGGER IF EXISTS decrease_folder_count_on_update");
        // 删除decrease_folder_count_on_delete触发器
        db.execSQL("DROP TRIGGER IF EXISTS decrease_folder_count_on_delete");
        // 删除delete_data_on_delete触发器
        db.execSQL("DROP TRIGGER IF EXISTS delete_data_on_delete");
        // 删除increase_folder_count_on_insert触发器
        db.execSQL("DROP TRIGGER IF EXISTS increase_folder_count_on_insert");
        // 删除folder_delete_notes_on_delete触发器
        db.execSQL("DROP TRIGGER IF EXISTS folder_delete_notes_on_delete");
        // 删除folder_move_notes_on_trash触发器
        db.execSQL("DROP TRIGGER IF EXISTS folder_move_notes_on_trash");

        // 创建increase_folder_count_on_update触发器
        db.execSQL(NOTE_INCREASE_FOLDER_COUNT_ON_UPDATE_TRIGGER);
        // 创建decrease_folder_count_on_update触发器
        db.execSQL(NOTE_DECREASE_FOLDER_COUNT_ON_UPDATE_TRIGGER);
        // 创建decrease_folder_count_on_delete触发器
        db.execSQL(NOTE_DECREASE_FOLDER_COUNT_ON_DELETE_TRIGGER);
        // 创建delete_data_on_delete触发器
        db.execSQL(NOTE_DELETE_DATA_ON_DELETE_TRIGGER);
        // 创建increase_folder_count_on_insert触发器
        db.execSQL(NOTE_INCREASE_FOLDER_COUNT_ON_INSERT_TRIGGER);
        // 创建folder_delete_notes_on_delete触发器
        db.execSQL(FOLDER_DELETE_NOTES_ON_DELETE_TRIGGER);
        // 创建folder_move_notes_on_trash触发器
        db.execSQL(FOLDER_MOVE_NOTES_ON_TRASH_TRIGGER);
    }

    /**
     * 创建系统文件夹
     */
    // 创建系统文件夹
    private void createSystemFolder(SQLiteDatabase db) {
        ContentValues values = new ContentValues();


        // 插入通话记录文件夹
        values.put(NoteColumns.ID, Notes.ID_CALL_RECORD_FOLDER);
        values.put(NoteColumns.TYPE, Notes.TYPE_SYSTEM);
        db.insert(TABLE.NOTE, null, values);


        // 清空values
        values.clear();
        // 插入根文件夹
        values.put(NoteColumns.ID, Notes.ID_ROOT_FOLDER);
        values.put(NoteColumns.TYPE, Notes.TYPE_SYSTEM);
        db.insert(TABLE.NOTE, null, values);

        /**
         * temporary folder which is used for moving note
         */
        values.clear();
        values.put(NoteColumns.ID, Notes.ID_TEMPARAY_FOLDER);
        values.put(NoteColumns.TYPE, Notes.TYPE_SYSTEM);
        db.insert(TABLE.NOTE, null, values);

        /**
         * create trash folder
         */
        values.clear();
        values.put(NoteColumns.ID, Notes.ID_TRASH_FOLER);
        values.put(NoteColumns.TYPE, Notes.TYPE_SYSTEM);
        db.insert(TABLE.NOTE, null, values);
    }

    /**
     * 创建data表
     */
    public void createDataTable(SQLiteDatabase db) {
        db.execSQL(CREATE_DATA_TABLE_SQL);
        reCreateDataTableTriggers(db);
        db.execSQL(CREATE_DATA_NOTE_ID_INDEX_SQL);
        Log.d(TAG, "data table has been created");
    }

    /**
     * 重新创建data表的触发器
     */
    private void reCreateDataTableTriggers(SQLiteDatabase db) {
        // 删除原有的触发器
        db.execSQL("DROP TRIGGER IF EXISTS update_note_content_on_insert");
        db.execSQL("DROP TRIGGER IF EXISTS update_note_content_on_update");
        db.execSQL("DROP TRIGGER IF EXISTS update_note_content_on_delete");

        // 创建新的触发器
        db.execSQL(DATA_UPDATE_NOTE_CONTENT_ON_INSERT_TRIGGER);
        db.execSQL(DATA_UPDATE_NOTE_CONTENT_ON_UPDATE_TRIGGER);
        db.execSQL(DATA_UPDATE_NOTE_CONTENT_ON_DELETE_TRIGGER);
    }

    /**
     * 获取NotesDatabaseHelper实例
     */
    // 获取NotesDatabaseHelper实例
    static synchronized NotesDatabaseHelper getInstance(Context context) {
        // 如果mInstance为空，则创建一个新的NotesDatabaseHelper实例
        if (mInstance == null) {
            mInstance = new NotesDatabaseHelper(context);
        }
        // 返回mInstance
        return mInstance;
    }

    // 创建数据库
    @Override
    public void onCreate(SQLiteDatabase db) {
        // 创建Note表
        createNoteTable(db);
        // 创建Data表
        createDataTable(db);
    }

    @Override
    public void onUpgrade(SQLiteDatabase db, int oldVersion, int newVersion) {
        // 定义两个布尔变量，用于判断是否需要重新创建触发器和是否跳过v2版本
        boolean reCreateTriggers = false;
        boolean skipV2 = false;

        if (oldVersion == 1) {
            upgradeToV2(db);
            skipV2 = true; // this upgrade including the upgrade from v2 to v3
            oldVersion++;
        }

        if (oldVersion == 2 && !skipV2) {
            upgradeToV3(db);
            reCreateTriggers = true;
            oldVersion++;
        }

        if (oldVersion == 3) {
            upgradeToV4(db);
            oldVersion++;
        }

        if (reCreateTriggers) {
            reCreateNoteTableTriggers(db);
            reCreateDataTableTriggers(db);
        }

        if (oldVersion != newVersion) {
            throw new IllegalStateException("Upgrade notes database to version " + newVersion
                    + "fails");
        }
    }

    /**
     * 升级到v2版本
     */
    private void upgradeToV2(SQLiteDatabase db) {
        db.execSQL("DROP TABLE IF EXISTS " + TABLE.NOTE);
        db.execSQL("DROP TABLE IF EXISTS " + TABLE.DATA);
        createNoteTable(db);
        createDataTable(db);
    }

    /**
     * 升级到v3版本
     */
    private void upgradeToV3(SQLiteDatabase db) {
        // drop unused triggers
        db.execSQL("DROP TRIGGER IF EXISTS update_note_modified_date_on_insert");
        db.execSQL("DROP TRIGGER IF EXISTS update_note_modified_date_on_delete");
        db.execSQL("DROP TRIGGER IF EXISTS update_note_modified_date_on_update");
        // add a column for gtask id
        db.execSQL("ALTER TABLE " + TABLE.NOTE + " ADD COLUMN " + NoteColumns.GTASK_ID
                + " TEXT NOT NULL DEFAULT ''");
        // add a trash system folder
        ContentValues values = new ContentValues();
        values.put(NoteColumns.ID, Notes.ID_TRASH_FOLER);
        values.put(NoteColumns.TYPE, Notes.TYPE_SYSTEM);
        db.insert(TABLE.NOTE, null, values);
    }

    /**
     * 升级到v4版本
     */
    private void upgradeToV4(SQLiteDatabase db) {
        db.execSQL("ALTER TABLE " + TABLE.NOTE + " ADD COLUMN " + NoteColumns.VERSION
                + " INTEGER NOT NULL DEFAULT 0");
    }
}
